import warnings

import pandas as pd
import plotly.express as px
import plotly.graph_objs as go

from ..statistical_testing import eval_ttest_significance, t_test
from .base import BaseEvaluator


class ABTestEvaluator(BaseEvaluator):
    def __init__(self) -> None:
        super().__init__()
        self.variant_col: str = ""

    def evaluate(
        self,
        data: pd.DataFrame,
        unit_col: str,
        metrics: list[str],
        variant_col: str = "variant",
    ) -> None:
        """calculate stats of A/B test and cache it into the class variable.
        At first, it only assumes metrics can handle by Welch's t-test.

        Parameters
        ----------
        data : pd.DataFrame
            Dataframe has randomization unit column, variant assignment column, and metrics columns.
            The data should have been aggregated by the randomization unit.
        unit_col : str
            A column name stores the randomization unit. something like user_id, session_id, ...
        variant_col : str
            A column name stores the variant assignment.
            The control variant should have value 1.
        metrics : list[str]
            Columns stores metrics you want to evaluate.
        """
        self.stats = t_test(data, unit_col, variant_col, metrics)
        self.variant_col = variant_col

    def summary_table(self, p_threshold: float = 0.05) -> pd.DataFrame:
        """return statistics summary.

        Parameters
        ----------
        p_threshold : float, optional
            significance level, by default 0.05

        Returns
        -------
        pd.DataFrame
            stats summary
        """
        self._validate_evaluate_executed()

        stats = self.stats.copy(deep=True)
        significance, abs_ci_width, rel_ci_width = eval_ttest_significance(self.stats, p_threshold)
        stats["significance"] = significance
        stats["abs_ci_width"] = abs_ci_width
        stats["rel_ci_width"] = rel_ci_width
        for diff_type in ["abs", "rel"]:
            stats[f"ci_{diff_type}_diff"] = stats.apply(
                lambda x: (
                    x[f"{diff_type}_diff_mean"] - x[f"{diff_type}_ci_width"],
                    x[f"{diff_type}_diff_mean"] + x[f"{diff_type}_ci_width"],
                ),
                axis=1,
            )
            del stats[f"{diff_type}_ci_width"]
        return_cols = [col for col in stats.columns if col[-2:] != "_c"]
        return stats.loc[:, return_cols]

    def summary_plot(self, p_threshold: float = 0.05, diff_type: str = "rel", display_ci: bool = True) -> go.Figure:
        """plot impact and confidence interval for each metric

        Parameters
        ----------
        p_threshold : float, optional
            significance level, by default 0.05
        diff_type : str, optional
            The type of difference you want to plot, by default 'rel'
        display_ci : bool, optional
            Whether to display confidence interval, by default True

        Returns
        -------
        go.Figure
        """
        self._validate_evaluate_executed()
        if diff_type not in ["rel", "abs"]:
            raise ValueError("Specified diff type is invalid.")

        stats = self.stats.copy(deep=True)
        significance, abs_ci_width, rel_ci_width = eval_ttest_significance(self.stats, p_threshold)
        stats["significant"] = significance
        stats["abs_ci_width"] = abs_ci_width
        stats["rel_ci_width"] = rel_ci_width

        viz_options = {
            "data_frame": stats.query(f"{self.variant_col} > 1"),  # not display control group
            "x": f"{diff_type}_diff_mean",
            "y": "metric",
            "facet_col": self.variant_col,
            "color": "significant",
            "color_discrete_map": {"up": "#54A24B", "down": "#E45756", "unclear": "silver"},
        }
        if display_ci:
            viz_options["error_x"] = f"{diff_type}_ci_width"

        g = px.bar(**viz_options)
        return g

    def summary_barplot(self, p_threshold: float = 0.05, diff_type: str = "rel", display_ci: bool = True) -> go.Figure:
        """plot impact and confidence interval for each metric

        Parameters
        ----------
        p_threshold : float, optional
            significance level, by default 0.05
        diff_type : str, optional
            The type of difference you want to plot, by default 'rel'
        display_ci : bool, optional
            Whether to display confidence interval, by default True

        Returns
        -------
        go.Figure
        """
        warnings.warn(
            "The summary_barplot() is deprecated and will be removed in the future release."
            "Please use summary_plot() instead.",
            FutureWarning,
        )
        return self.summary_plot(p_threshold=p_threshold, diff_type=diff_type, display_ci=display_ci)



import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objs as go
from scipy.stats import kstest

from ..statistical_testing import t_test
from .base import BaseEvaluator


class AATestEvaluator(BaseEvaluator):
    def __init__(self, n_simulation: int = 1000, sample_rate: float = 1.0) -> None:
        """initialize parameters affect result of evaluation.

        Parameters
        ----------
        n_simulation : float, optional
            How many times you want to do simulated A/A test, by default 1000
        sample_rate : float, optional
            How much fraction you want to sample from the dataframe, by default 1.0
        """
        if n_simulation <= 0:
            raise ValueError("The number of simulation should be positive number.")
        if not 0.0 < sample_rate <= 1.0:
            raise ValueError("The sample rate should be in (0, 1]")

        super().__init__()
        self.n_simulation = n_simulation
        self.sample_rate = sample_rate

    def evaluate(self, data: pd.DataFrame, unit_col: str, metrics: list[str]) -> None:
        """split data n times, and calculate statistics n times, then store it as an attribute.

        Parameters
        ----------
        data : pd.DataFrame
            Dataframe has randomization unit column, and metrics columns. The data should have been aggregated by the randomization unit.
        unit_col : str
            A column name stores the randomization unit. something like user_id, session_id, ...
        metrics : list[str]
            Columns stores metrics you want to evaluate.
        """
        result = pd.DataFrame()
        sampled_df = data.copy(deep=True)
        for i in range(self.n_simulation):
            if self.sample_rate < 1.0:
                sampled_df = data.sample(frac=self.sample_rate, random_state=i, ignore_index=True)
            sampled_df["variant"] = np.random.randint(
                low=1, high=3, size=sampled_df.shape[0]
            )  # The upper bound is exclusive
            stats = t_test(sampled_df, unit_col, "variant", metrics)
            stats["idx"] = i
            result = pd.concat([result, stats])
        return_cols = [
            "idx",
            "metric",
            "mean",
            "count",
            "stderr",
            "abs_diff_mean",
            "abs_diff_std",
            "rel_diff_mean",
            "p_value",
        ]
        result = result.query("variant == 2").loc[:, return_cols].reset_index(drop=True)
        self.stats = result

    def summary_table(self) -> pd.DataFrame:
        """Apply Kolmogorov Smirnov test to check if the p-value distribution is different from the uniform distribution.

        Returns
        -------
        pd.DataFrame
            stats summary
        """
        self._validate_evaluate_executed()

        stats_agg = (
            self.stats.groupby("metric")[["abs_diff_mean", "rel_diff_mean", "p_value"]]
            .agg(["mean", "std"])
            .reset_index()
        )
        stats_agg["ksstat"] = 0
        stats_agg["ks_pvalue"] = 0
        for metric in self.stats["metric"].unique():
            result = kstest(self.stats.query(f"metric == '{metric}'")["p_value"], "uniform")
            stats_agg.loc[stats_agg["metric"] == metric, "ksstat"] = result.statistic
            stats_agg.loc[stats_agg["metric"] == metric, "ks_pvalue"] = result.pvalue
        stats_agg["significance"] = stats_agg["ks_pvalue"].map(lambda x: True if x < 0.05 else False)
        return stats_agg

    def summary_plot(self) -> go.Figure:
        """Plot histogram of p-value with coloring if the p-value distribution is different from the uniform distribution.

        Returns
        -------
        go.Figure
        """
        self._validate_evaluate_executed()

        stats = self.stats.copy(deep=True)
        stats["ks_pvalue"] = 0
        for metric in stats["metric"].unique():
            result = kstest(stats.query(f"metric == '{metric}'")["p_value"], "uniform")
            stats.loc[stats["metric"] == metric, "ks_pvalue"] = result.pvalue
        stats["significance"] = stats["ks_pvalue"].map(lambda x: True if x < 0.05 else False)

        g = px.histogram(
            data_frame=stats,
            x="p_value",
            facet_col="metric",
            nbins=100,
            histnorm="probability",
            color="significance",
            color_discrete_map={False: "#54A24B", True: "#E45756"},
        )
        g.add_hline(y=0.01, line_dash="dot", line_width=2)
        return g



from abc import ABC, abstractmethod

import pandas as pd
import plotly.graph_objs as go


class BaseEvaluator(ABC):
    def __init__(self) -> None:
        self.stats: pd.DataFrame = pd.DataFrame()

    @abstractmethod
    def evaluate(self, data: pd.DataFrame, unit_col: str, metrics: list[str]) -> None:
        pass

    @abstractmethod
    def summary_table(self) -> pd.DataFrame:
        return pd.DataFrame()

    @abstractmethod
    def summary_plot(self) -> go.Figure:
        return go.Figure()

    def _validate_evaluate_executed(self) -> None:
        if self.stats.shape[0] == 0:
            raise ValueError("Evaluated statistics haven't been calculated. Please call evaluate() in advance.")



import numpy as np
import pandas as pd
from scipy.stats import t, ttest_ind_from_stats


def t_test(data: pd.DataFrame, unit_col: str, variant_col: str, metrics: list[str]) -> pd.DataFrame:
    """_summary_

    Parameters
    ----------
    data : pd.DataFrame
        A DataFrame has randomization unit column, variant assignment column, and metrics columns.
        The data should have been aggregated by the randomization unit.
    unit_col : str
        A column name stores the randomization unit. something like user_id, session_id, ...
    variant_col : str
        A column name stores the variant assignment.
        The control variant should have value 1.
    metrics : list[str]
        Columns stores metrics you want to evaluate.

    Returns
    -------
    pd.DataFrame
        A DataFrame stores basic statistics of each variant (mean, variance, ...) and impact compared with control group.
    """
    if data.shape[0] != data[unit_col].nunique():
        raise ValueError("passed dataframe hasn't been aggregated by the randomization unit.")
    if len(metrics) == 0:
        raise ValueError("metrics hasn't been specified.")
    if data[variant_col].min() != 1:
        raise ValueError("the control variant seems not to exist.")
    means = (
        data.groupby(variant_col)[metrics].mean().stack().reset_index().rename(columns={"level_1": "metric", 0: "mean"})
    )
    vars = (
        data.groupby(variant_col)[metrics].var().stack().reset_index().rename(columns={"level_1": "metric", 0: "var"})
    )
    counts = (
        data.groupby(variant_col)[metrics]
        .count()
        .stack()
        .reset_index()
        .rename(columns={"level_1": "metric", 0: "count"})
    )
    stats = means.merge(vars, on=[variant_col, "metric"]).merge(counts, on=[variant_col, "metric"])
    stats["std"] = np.sqrt(stats["var"])
    stats["stderr"] = np.sqrt(stats["var"] / stats["count"])
    stats = stats.merge(stats.query(f"{variant_col} == 1"), on="metric", suffixes=["", "_c"]).drop("variant_c", axis=1)

    stats["abs_diff_mean"] = stats["mean"] - stats["mean_c"]
    stats["abs_diff_std"] = np.sqrt(stats["stderr"] ** 2 + stats["stderr_c"] ** 2)
    stats["rel_diff_mean"] = stats["mean"] / stats["mean_c"] - 1

    # see: https://arxiv.org/pdf/1803.06336.pdf
    stats["rel_diff_std"] = (
        np.sqrt((stats["stderr"] ** 2 + (stats["mean"] ** 2 / stats["mean_c"] ** 2) * stats["stderr_c"] ** 2))
        / stats["mean_c"]
    )
    stats["t_value"] = stats["abs_diff_mean"] / stats["abs_diff_std"]
    stats["dof"] = stats["abs_diff_std"] ** 4 / (
        stats["stderr"] ** 4 / (stats["count"] - 1) + stats["stderr_c"] ** 4 / (stats["count_c"] - 1)
    )
    stats["p_value"] = stats.apply(
        lambda x: ttest_ind_from_stats(
            mean1=x["mean"],
            std1=x["std"],
            nobs1=x["count"],
            mean2=x["mean_c"],
            std2=x["std_c"],
            nobs2=x["count_c"],
            equal_var=False,
        ).pvalue,
        axis=1,
    )
    return stats


def eval_ttest_significance(
    ttest_stats: pd.DataFrame, p_threshold: float = 0.05
) -> tuple[pd.Series, pd.Series, pd.Series]:
    """evaluate statistical significance & return information related to that.

    Parameters
    ----------
    ttest_stats : pd.DataFrame
        calculated statistics in t_test() function.
    p_threshold : float, optional
        significant level, by default 0.05

    Returns
    -------
    tuple[pd.Series]
        includes three pd.Series as follows:
        - statistical significance
        - confidence interval width of absolute difference
        - confidence interval width of relative difference
    """
    necessary_cols = ["p_value", "t_value", "dof", "abs_diff_std", "rel_diff_std"]
    for col in necessary_cols:
        if col in ttest_stats.columns:
            continue
        raise ValueError(f"Necessary column does not exist: {col}")

    significance = ttest_stats.apply(
        lambda x: "up"
        if x["p_value"] <= p_threshold and x["t_value"] > 0
        else "down"
        if x["p_value"] <= p_threshold and x["t_value"] < 0
        else "unclear",
        axis=1,
    )

    abs_ci_width = t.ppf(1 - p_threshold / 2, ttest_stats["dof"]) * ttest_stats["abs_diff_std"]
    rel_ci_width = t.ppf(1 - p_threshold / 2, ttest_stats["dof"]) * ttest_stats["rel_diff_std"]

    return significance, abs_ci_width, rel_ci_width
